
#include <cpse/math/matrix_assign_elements.tpp>

namespace cpse {
namespace math {

// assign the "common" part
template<int A, int B, int C, int D> inline
void subassign(basic_matrix<A, B> &dest, const basic_matrix<C, D> &m)
{
	const int E = A < B ? A : B;
	const int F = C < D ? C : D;
	for (int i=0; i!=E; ++i) for (int j=0; j!=F; ++j)
		dest(i,j) = m(i,j);
}

// zero
template<int A, int B> inline
void assign(basic_matrix<A, B> &dest, zero_t)
	{ for (float *p = dest.data(); p != dest.data()+A*B; ++p) *p = 0.f; }


// expression dispatch

struct erroneous_expression_handler {};
struct scalar_assign_handler;
struct matrix_assign_handler;
struct matrix2_assign_handler;
struct matrix3_assign_handler;
struct matrix4_assign_handler;

template<int R, int C, class S, int SR, int SC>
struct assign_dispatch_3_elop {
	// size mismatch
	typedef erroneous_expression_handler handler;
};
template<int R, int C, class S>
struct assign_dispatch_3_elop<R, C, S, R, C> {
	typedef assign_elop_handler<R*C> handler;
};

template<class D, class S, bool ElOp>
struct assign_dispatch_2 {
	// generic
	typedef matrix_assign_handler handler;
};
template<int R, int C, class S>
struct assign_dispatch_2<basic_matrix<R, C>, S, true> {
	// element_op
	typedef typename assign_dispatch_3_elop<R, C, S, S::rows,
		S::columns>::handler handler;
};
template<int C, class S>
struct assign_dispatch_2<basic_matrix<2, C>, S, false> {
	// matrix2 result
	typedef matrix2_assign_handler handler;
};
template<int C, class S>
struct assign_dispatch_2<basic_matrix<3, C>, S, false> {
	// matrix3 result
	typedef matrix3_assign_handler handler;
};
template<int C, class S>
struct assign_dispatch_2<basic_matrix<4, C>, S, false> {
	// matrix4 result
	typedef matrix4_assign_handler handler;
};

template<class D, class S, bool Error, bool Scalar>
struct assign_dispatch_1 { // detect errors
	typedef erroneous_expression_handler handler;
};
template<class D, class S>
struct assign_dispatch_1<D, S, false, true> {  // scalar result
	typedef scalar_assign_handler handler;
};
template<class D, class S>
struct assign_dispatch_1<D, S, false, false> { // matrix result
	typedef typename assign_dispatch_2<D, S, S::element_op>::handler 
		handler;
};
template<int R, int C>
struct assign_dispatch_1<basic_matrix<R, C>, basic_matrix<R, C>, false, false> {
	// copy
	struct handler {
		static void go(basic_matrix<R, C> &dest,
			const basic_matrix<R, C> &src) { dest = src; }
	};
};

template<class D, class S> struct assign_dispatch {
	typedef typename assign_dispatch_1<typename D::base_type,
		typename S::base_type, S::error
			|| D::is_matrix && !S::is_matrix
			|| D::is_scalar && !S::is_scalar, 
			D::is_scalar>::handler handler;
};

struct scalar_assign_handler {
	template<class Op, class T>
	static void go(float_keeper &d, matrix_unop<Op, T> s) {
		typedef matrix_unop<Op, typename result_type<typename 
			T::base_type>::scalar_actual> combined;
		typedef typename assign_dispatch<float_keeper, 
			combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	template<class Op, class T, class U>
	static void go(float_keeper &d, matrix_binop<Op, T, U> s) {
		typedef typename result_type<typename 
			T::base_type>::scalar_actual TA;
		typedef typename result_type<typename 
			U::base_type>::scalar_actual UA;
		typedef matrix_binop<Op, TA, UA> combined;
		typedef typename assign_dispatch<float_keeper, 
			combined>::handler handler;
		handler::go(d, combined(s.a, s.b));
	}

	static void go(float_keeper &d, matrix_unop<unop_negate,
		float_keeper> s) { d.f = -s.a.f; }

	static void go(float_keeper &d, matrix_binop<binop_mul, 
		float_keeper, float_keeper> s) { d.f = s.a.f * s.b.f; }
	static void go(float_keeper &d, matrix_binop<binop_div, 
		float_keeper, float_keeper> s) { d.f = s.a.f / s.b.f; }
	static void go(float_keeper &d, matrix_binop<binop_add, 
		float_keeper, float_keeper> s) { d.f = s.a.f + s.b.f; }
	static void go(float_keeper &d, matrix_binop<binop_sub, 
		float_keeper, float_keeper> s) { d.f = s.a.f - s.b.f; }

	// scalar (dot) product
	static void go(float_keeper &dest, matrix_binop<binop_mul, 
		basic_matrix<2, 1>, basic_matrix<2, 1> > s)
			{ dest.f = s.a.x*s.b.x + s.a.y*s.b.y; }
	static void go(float_keeper &dest, matrix_binop<binop_mul, 
		basic_matrix<3, 1>, basic_matrix<3, 1> > s)
			{ dest.f = s.a.x*s.b.x + s.a.y*s.b.y + s.a.z*s.b.z; }
	static void go(float_keeper &dest, matrix_binop<binop_mul, 
		basic_matrix<4, 1>, basic_matrix<4, 1> > s) {
		dest.f = s.a.x*s.b.x + s.a.y*s.b.y + s.a.z*s.b.z + s.a.w*s.b.w;
	}
};

struct matrix_assign_handler {

	template<int R, int C, class Op, class T>
	static void go(basic_matrix<R, C> &d, matrix_unop<Op, T> s) {
		typedef matrix_unop<Op, typename result_type<typename 
			T::base_type>::matrix_actual> combined;
		typedef typename assign_dispatch<basic_matrix<R, C>, 
			combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	template<int R, int C, class Op, class T, class U>
	static void go(basic_matrix<R, C> &d, matrix_binop<Op, T, U> s) {
		typedef typename result_type<typename 
			T::base_type>::matrix_actual TA;
		typedef typename result_type<typename 
			U::base_type>::matrix_actual UA;
		typedef matrix_binop<Op, TA, UA> combined;
		typedef typename assign_dispatch<basic_matrix<R, C>, 
			combined>::handler handler;
		handler::go(d, combined(s.a, s.b));
	}

	// assign the matching part of matrix
	template<int A, int B, int C, int D>
	static void go(basic_matrix<A, B> &dest, matrix_unop<unop_slice,
		basic_matrix<C, D> > src) { subassign(dest, src.a); }

	// generic matrix multiplication
	template<int A, int C>
	static void go(basic_matrix<A, C> &, matrix_binop<binop_mul,
		basic_matrix<A, 2>, basic_matrix<2, C> >);
	template<int A, int C>
	static void go(basic_matrix<A, C> &, matrix_binop<binop_mul,
		basic_matrix<A, 3>, basic_matrix<3, C> >);
	template<int A, int C>
	static void go(basic_matrix<A, C> &, matrix_binop<binop_mul,
		basic_matrix<A, 4>, basic_matrix<4, C> >);

};

// matrix multiplication
template<int A, int C>
void matrix_assign_handler::go(basic_matrix<A, C> &d, matrix_binop<binop_mul,
	basic_matrix<A, 2>, basic_matrix<2, C> > s)
{
	for (int i=0; i!=A; ++i) for (int j=0; j!=C; ++j)
		d(i,j) = s.a(i,0)*s.b(0,j) + s.a(i,1)*s.b(1,j);
}
template<int A, int C>
void matrix_assign_handler::go(basic_matrix<A, C> &d, matrix_binop<binop_mul,
	basic_matrix<A, 3>, basic_matrix<3, C> > s)
{
	for (int i=0; i!=A; ++i) for (int j=0; j!=C; ++j)
		d(i,j) = s.a(i,0)*s.b(0,j) + s.a(i,1)*s.b(1,j)
			+ s.a(i,2)*s.b(2,j);
}
template<int A, int C>
void matrix_assign_handler::go(basic_matrix<A, C> &d, matrix_binop<binop_mul,
	basic_matrix<A, 4>, basic_matrix<4, C> > s)
{
	for (int i=0; i!=A; ++i) for (int j=0; j!=C; ++j)
		d(i,j) = s.a(i,0)*s.b(0,j) + s.a(i,1)*s.b(1,j)
			+ s.a(i,2)*s.b(2,j) + s.a(i,3)*s.b(3,j);
}

// more convenient than assign_dispatch, but potentially less efficient
template<class D, class S> inline
void assign_matrix(D &dest, const S &src) {
	typedef typename assign_dispatch<D, S>::handler handler;
	handler::go(dest, src);
}

} // namespace math
} // namespace cpse

#include <cpse/math/matrix_assign_2.tpp>
#include <cpse/math/matrix_assign_3.tpp>
#include <cpse/math/matrix_assign_4.tpp>
