
namespace cpse {
namespace math {

template<class Op, class A, class B> struct matrix_binop {};

template<class A, class B>
struct matrix_binop<binop_mul, A, B> {
	typedef typename select_matrix<A, B>::type possibly_A;
	typedef typename select_matrix<B, A>::type possibly_B;
	static const int rows = possibly_A::rows;
	static const int columns = possibly_B::columns;
	enum { matrix_mul = A::columns == B::rows && A::is_matrix
		&& B::is_matrix };
	enum { matrix_scalar_mul = A::is_matrix && B::is_scalar
		|| A::is_scalar && B::is_matrix };
	enum { vector_product = is_vector_3<A>::value
		&& is_vector_3<B>::value };
	enum { scalar_product = A::columns == 1 && B::columns == 1
		&& A::rows == B::rows && A::is_matrix && B::is_matrix };
	enum { scalar_mul = A::is_scalar && B::is_scalar };
	enum { product_ambiguity = vector_product && matrix_scalar_mul };
	static const bool is_matrix = matrix_mul || vector_product
		|| matrix_scalar_mul;
	static const bool is_scalar = scalar_product || scalar_mul;
	static const bool error = !is_matrix && !is_scalar || A::error
		|| B::error || product_ambiguity;
	static const bool element_op = A::element_op && B::element_op
		&& !matrix_mul && !vector_product && !scalar_product;
	typedef matrix_binop<binop_mul, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int I> float element() const
		{ return a.element<I>() * b.element<I>(); }
};

template<class A, class B>
struct matrix_binop<binop_cross, A, B> {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_matrix = is_vector_3<A>::value
		&& is_vector_3<B>::value;
	static const bool is_scalar = false;
	static const bool error = A::error || B::error || !is_matrix;
	static const bool element_op = false;
	typedef matrix_binop<binop_cross, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
};

template<class A, class B>
struct matrix_binop<binop_div, A, B> {
	typedef select_matrix<A, B>::type possibly_A;
	static const int rows = possibly_A::rows;
	static const int columns = possibly_A::columns;
	enum { matrix_scalar = A::is_matrix && B::is_scalar };
	enum { scalar_scalar = A::is_scalar && B::is_scalar };
	static const bool is_matrix = matrix_scalar;
	static const bool is_scalar = scalar_scalar;
	static const bool error = A::error || B::error || !is_matrix
		&& !is_scalar;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_div, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int I> float element() const
		{ return a.element<I>() / b.element<I>(); }
};

template<class A, class B>
struct matrix_binop<binop_add, A, B> {
	static const int rows = select_matrix<A, B>::type::rows;
	static const int columns = select_matrix<A, B>::type::columns;
	static const bool is_matrix = A::is_matrix && B::is_matrix;
	static const bool is_scalar = A::is_scalar && B::is_scalar;
	static const bool error = A::error || B::error
		|| !is_matrix && !is_scalar;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_add, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int I> float element() const
		{ return a.element<I>() + b.element<I>(); }
};

template<class A, class B>
struct matrix_binop<binop_sub, A, B> {
	static const int rows = select_matrix<A, B>::type::rows;
	static const int columns = select_matrix<A, B>::type::columns;
	static const bool is_matrix = A::is_matrix && B::is_matrix;
	static const bool is_scalar = A::is_scalar && B::is_scalar;
	static const bool error = A::error || B::error
		|| !is_matrix && !is_scalar;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_sub, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int I> float element() const
		{ return a.element<I>() - b.element<I>(); }
};

template<class A, class B>
struct matrix_binop<binop_expand, A, B> {
	static const int rows = A::rows + (B::is_matrix ? 2 : 1);
	static const int columns = 1;
	static const bool is_matrix = A::is_matrix
		&& (B::is_scalar || B::is_matrix);
	static const bool is_scalar = false;
	static const bool error = !is_matrix || A::error;
	static const bool element_op = false;
	typedef matrix_binop<binop_expand, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
};

template<class A, class B>
struct matrix_binop<binop_rotation, A, B> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_matrix = true;
	static const bool is_scalar = false;
	static const bool error = !A::is_matrix || A::error || B::error
		|| !B::is_scalar;
	static const bool element_op = false;
	typedef matrix_binop<binop_rotation, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
};

} // namespace math
} // namespace cpse
